
/*
 *    ChunkSync - create space-efficient incremental (remote) backups of large files
 *    Copyright (C) 2009  Florian Zumbiehl <florz@florz.de>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 *    In addition, as a special exception, the author of this
 *    program gives permission to link the code of its 
 *    release with the OpenSSL project's "OpenSSL" library (or
 *    with modified versions of it that use the same license as
 *    the "OpenSSL" library), and distribute the linked
 *    executables. You must obey the GNU General Public 
 *    License in all respects for all of the code used other 
 *    than "OpenSSL".  If you modify this file, you may extend
 *    this exception to your version of the file, but you are
 *    not obligated to do so.  If you do not wish to do so,
 *    delete this exception statement from your version.
 */

#define _GNU_SOURCE

#include <stdbool.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <pthread.h>

#include "progress.h"
#include "err.h"

static const char *progname;
static bool backend_mode;
static pthread_mutex_t lock=PTHREAD_ERRORCHECK_MUTEX_INITIALIZER_NP;

void err_set_progname(const char *name){
	progname=name;
}

void err_set_backend_mode(bool mode){
	backend_mode=mode;
}

void err_start(bool progress){
	if(pthread_mutex_lock(&lock))exit(1);
	if(!progress)progress_clear();
}

void err_end(){
	pthread_mutex_unlock(&lock);
}

void err_printname(const char *role){
	if(fprintf(stderr,"%s: %s: ",progname,role)<0)die(true,"fprintf(stderr)");
}

static void err_print(bool syserr,int my_errno,const char *fmt,va_list ap){
	if(!backend_mode)err_printname("frontend");
	if(vfprintf(stderr,fmt,ap)<0)die(true,"vfprintf(stderr)");
	if(syserr)if(fprintf(stderr,": %s",strerror(my_errno))<0)die(true,"fprintf(stderr)");
	if(fputc('\n',stderr)==EOF)die(true,"fputc(stderr)");
}

void warn(bool syserr,const char *fmt,...){
	int my_errno;
	va_list ap;

	my_errno=errno;
	va_start(ap,fmt);
	err_start(false);
	err_print(syserr,my_errno,fmt,ap);
	err_end();
	va_end(ap);
}

void die(bool syserr,const char *fmt,...){
	int my_errno;
	va_list ap;

	my_errno=errno;
	va_start(ap,fmt);
	err_start(false);
	err_print(syserr,my_errno,fmt,ap);
	exit(1);
}

void proterr(){
	die(false,"protocol error");
}

void usage(bool error){
	(error?&die:&warn)(false,"Usage: %s { -V | -h | [ -l <linkdir> [ ... ] ] <chunk size> <source> <destination> }",progname);
}

