
/*
 *    ChunkSync - create space-efficient incremental (remote) backups of large files
 *    Copyright (C) 2009  Florian Zumbiehl <florz@florz.de>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 *    In addition, as a special exception, the author of this
 *    program gives permission to link the code of its 
 *    release with the OpenSSL project's "OpenSSL" library (or
 *    with modified versions of it that use the same license as
 *    the "OpenSSL" library), and distribute the linked
 *    executables. You must obey the GNU General Public 
 *    License in all respects for all of the code used other 
 *    than "OpenSSL".  If you modify this file, you may extend
 *    this exception to your version of the file, but you are
 *    not obligated to do so.  If you do not wish to do so,
 *    delete this exception statement from your version.
 */

#define _GNU_SOURCE

#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <errno.h>
#include <pthread.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>
#include <arpa/inet.h>

#include "utils.h"
#include "err.h"
#include "io.h"

static void graceful_start(const struct xfh *fh){
	if(fh->graceful){
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
		fh->graceful(false);
	}
}

static void graceful_end(const struct xfh *fh){
	if(fh->graceful)pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
}

static void graceful_die(bool graceful,const struct xfh *fh,bool syserr,const char *func){
	graceful_end(fh);
	(fh->graceful&&graceful?&warn:&die)(syserr,"%s(%s)%s",func,fh->name,syserr?"":": Unexpected end of input");
	fh->graceful(true);
}

static inline uint64_t htonll(uint64_t in){
	uint64_t ret;

	for(size_t x=sizeof(ret);x--;){
		((unsigned char *)&ret)[x]=in;
		in>>=CHAR_BIT;
	}
	return ret;
}

static inline uint64_t ntohll(uint64_t in){
	uint64_t ret=0;

	for(size_t x=0;x<sizeof(ret);x++)
		ret=(ret<<CHAR_BIT)|((unsigned char *)&in)[x];
	return ret;
}

void xfh_init(struct xfh *fh,int r,int w,void (*graceful)(bool),const char *name){
	fh->r=r;
	fh->w=w;
	fh->graceful=graceful;
	fh->name=name;
}

void xfh_close(const struct xfh *fh){
	if(close(fh->r))die(true,"close(%s(r))",fh->name);
	if(close(fh->w))die(true,"close(%s(w))",fh->name);
}

void xwrite(const struct xfh *fh,const void *buf,size_t count){
	size_t done;
	ssize_t r;

	graceful_start(fh);
	for(done=0;done!=count;done+=r)
		if((r=write(fh->w,buf+done,count-done))<1)graceful_die(r&&errno==EPIPE,fh,r,"write");
	graceful_end(fh);
}

void xwrite_u8(const struct xfh *fh,uint8_t x){
	xwrite(fh,&x,1);
}

void xwrite_u32(const struct xfh *fh,uint32_t x){
	x=htonl(x);
	xwrite(fh,&x,sizeof(x));
}

void xwrite_u64(const struct xfh *fh,uint64_t x){
	x=htonll(x);
	xwrite(fh,&x,sizeof(x));
}

void xwrite_string(const struct xfh *fh,const char *s){
	size_t l;

	l=strlen(s);
	if(l>UINT32_MAX)die(false,"string to %s too long",fh->name);
	xwrite_u32(fh,l);
	xwrite(fh,s,l);
}

void xread(const struct xfh *fh,void *buf,size_t count){
	size_t done;
	ssize_t r;

	graceful_start(fh);
	for(done=0;done!=count;done+=r)
		if((r=read(fh->r,buf+done,count-done))<1)graceful_die(!r,fh,r,"read");
	graceful_end(fh);
}

uint8_t xread_u8(const struct xfh *fh){
	uint8_t x;

	xread(fh,&x,1);
	return x;
}

uint32_t xread_u32(const struct xfh *fh){
	uint32_t x;

	xread(fh,&x,sizeof(x));
	return ntohl(x);
}

uint64_t xread_u64(const struct xfh *fh){
	uint64_t x;

	xread(fh,&x,sizeof(x));
	return ntohll(x);
}

char *xread_string(const struct xfh *fh){
	uint32_t l;
	char *s;

	l=xread_u32(fh);
	if(l>=SIZE_MAX)die(false,"string size from %s too large",fh->name);
	if(!(s=malloc(((size_t)l)+1)))die(true,"malloc(string from %s)",fh->name);
	xread(fh,s,l);
	s[l]=0;
	return s;
}

void xcopy(const struct xfh *dest,const struct xfh *src,off_t count){
	char buf[64*1024];
	ssize_t r;

	while(count){
		graceful_start(src);
		if((r=read(src->r,buf,min(sizeof(buf),count)))<1)graceful_die(!r,src,r,"read");
		graceful_end(src);
		xwrite(dest,buf,r);
		count-=r;
	}
}

