
/*
 *    ChunkSync - create space-efficient incremental (remote) backups of large files
 *    Copyright (C) 2009  Florian Zumbiehl <florz@florz.de>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 *    In addition, as a special exception, the author of this
 *    program gives permission to link the code of its 
 *    release with the OpenSSL project's "OpenSSL" library (or
 *    with modified versions of it that use the same license as
 *    the "OpenSSL" library), and distribute the linked
 *    executables. You must obey the GNU General Public 
 *    License in all respects for all of the code used other 
 *    than "OpenSSL".  If you modify this file, you may extend
 *    this exception to your version of the file, but you are
 *    not obligated to do so.  If you do not wish to do so,
 *    delete this exception statement from your version.
 */

#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <arpa/inet.h>
#include <linux/fs.h>

#include <openssl/evp.h>

#include "utils.h"
#include "defs.h"
#include "io.h"
#include "backend.h"
#include "err.h"
#include "frontend.h"
#include "progress.h"
#include "sender.h"

static void dohash(EVP_MD_CTX *digest_ctx,unsigned char *chunk_digest,int image_fd,off_t this_chunk_size){
	char buf[64*1024];
	ssize_t r;

	if(!EVP_DigestInit_ex(digest_ctx,EVP_sha1(),NULL))die(false,"EVP_DigestInit_ex()");
	if(!EVP_DigestUpdate(digest_ctx,chunk_digest,DIGEST_SALT_SIZE))die(false,"EVP_DigestUpdate()");
	for(off_t count=this_chunk_size;count;count-=r){
		if((r=read(image_fd,buf,min(sizeof(buf),count)))<1)die(r<0,"read(image)");
		if(!EVP_DigestUpdate(digest_ctx,buf,r))die(false,"EVP_DigestUpdate()");
	}
	if(!EVP_DigestFinal_ex(digest_ctx,chunk_digest+DIGEST_SALT_SIZE,NULL))die(false,"EVP_DigestFinal_ex()");
}

void sender_main(const char *image_name){
	size_t linkdirs_cnt,linkdir_num,chunk_digests_cnt;
	uint64_t remote_chunk_size;
	off_t chunk_size,image_size,image_chunks,chunk_num,this_chunk_size;
	int image_fd,random_fd;
	struct xfh image_fh,random_fh;
	struct stat st;
	unsigned char chunk_digests[LINKDIRS_MAX][DIGEST_SALT_SIZE+DIGEST_SIZE],(*chunk_digest)[DIGEST_SALT_SIZE+DIGEST_SIZE]=NULL,remote_chunk_digest[DIGEST_SALT_SIZE+DIGEST_SIZE];
	EVP_MD_CTX digest_ctx;
	bool linkdirs_eof[LINKDIRS_MAX],chunk_done;
	uint8_t transfer;

	xwrite_u8(&stdio_fh,1);
	if((linkdirs_cnt=xread_u8(&stdio_fh))>LINKDIRS_MAX)proterr();
	remote_chunk_size=xread_u64(&stdio_fh);
	if(remote_chunk_size<1||remote_chunk_size>ANYSINT_MAX(off_t))proterr();
	chunk_size=remote_chunk_size;
	if((image_fd=open(image_name,O_RDONLY))<0)die(true,"open(source)");
	xfh_init(&image_fh,image_fd,-1,NULL,"source");
	if((errno=posix_fadvise(image_fd,0,0,POSIX_FADV_SEQUENTIAL)))die(true,"posix_fadvise()");
	if(fstat(image_fd,&st)<0)die(true,"stat(source)");
	if(S_ISBLK(st.st_mode)){
		uint64_t blksize;
		if(ioctl(image_fd,BLKGETSIZE64,&blksize)<0)die(true,"ioctl(source,BLKGETSIZE64)");
		if(blksize>ANYSINT_MAX(off_t))die(false,"block device too large");
		image_size=blksize;
	}else{
		image_size=st.st_size;
	}
	image_chunks=image_size/chunk_size;
	if(image_size%chunk_size)image_chunks++;
	xwrite_u64(&stdio_fh,image_chunks);
	if((random_fd=open("/dev/urandom",O_RDONLY))<0)die(true,"open(/dev/urandom)");
	xfh_init(&random_fh,random_fd,-1,NULL,"/dev/urandom");
	EVP_MD_CTX_init(&digest_ctx);
	memset(linkdirs_eof,0,sizeof(linkdirs_eof));
	remote_chunk_size=1;
	memset(remote_chunk_digest,0,sizeof(remote_chunk_digest));
	this_chunk_size=chunk_size;
	for(chunk_num=0;chunk_num<image_chunks;chunk_num++){
		if(chunk_num==image_chunks-1&&image_size%chunk_size)this_chunk_size=image_size%chunk_size;
		chunk_done=false;
		chunk_digests_cnt=0;
		for(linkdir_num=0;linkdir_num<linkdirs_cnt;linkdir_num++)if(!linkdirs_eof[linkdir_num])
			switch((transfer=xread_u8(&stdio_fh))>>3){
				case 0:
					if(transfer&7)proterr();
					linkdirs_eof[linkdir_num]=true;
					break;
				case 1:
					if(transfer&1){
						remote_chunk_size=xread_u64(&stdio_fh);
						if(remote_chunk_size<1||remote_chunk_size>ANYSINT_MAX(off_t))proterr();
					}
					if(transfer&2)xread(&stdio_fh,remote_chunk_digest,DIGEST_SALT_SIZE);
					if(transfer&4)xread(&stdio_fh,remote_chunk_digest+DIGEST_SALT_SIZE,DIGEST_SIZE);
					if(!chunk_done){
						if(!chunk_digests_cnt||transfer&2){
							for(chunk_digest=chunk_digests;chunk_digest<chunk_digests+chunk_digests_cnt&&memcmp(chunk_digest,remote_chunk_digest,DIGEST_SALT_SIZE);chunk_digest++);
							if(chunk_digest==chunk_digests+chunk_digests_cnt){
								if(lseek(image_fd,chunk_num*chunk_size,SEEK_SET)<0)die(true,"lseek()");
								memcpy(chunk_digest,remote_chunk_digest,DIGEST_SALT_SIZE);
								dohash(&digest_ctx,*chunk_digest,image_fd,this_chunk_size);
								chunk_digests_cnt++;
							}
						}
						if(remote_chunk_size==this_chunk_size&&!memcmp(remote_chunk_digest+DIGEST_SALT_SIZE,(*chunk_digest)+DIGEST_SALT_SIZE,DIGEST_SIZE)){
							xwrite_u8(&stdio_fh,(0<<6)|linkdir_num);
							chunk_done=true;
						}
					}
					break;
				default:
					proterr();
			}
		if(!chunk_done){
			if(!chunk_digests_cnt){
				chunk_digest=chunk_digests;
				xread(&random_fh,chunk_digest,DIGEST_SALT_SIZE);
				dohash(&digest_ctx,*chunk_digest,image_fd,this_chunk_size);
			}
			xwrite_u8(&stdio_fh,1<<6);
			xwrite_u64(&stdio_fh,this_chunk_size);
			xwrite(&stdio_fh,chunk_digest,DIGEST_SALT_SIZE+DIGEST_SIZE);
			if(lseek(image_fd,-this_chunk_size,SEEK_CUR)<0)die(true,"lseek()");
			xcopy(&stdio_fh,&image_fh,this_chunk_size);
		}
	}
	EVP_MD_CTX_cleanup(&digest_ctx);
	if(close(random_fd))die(true,"close(/dev/urandom)");
	if(close(image_fd))die(true,"close(source)");
	xwrite_u8(&stdio_fh,2<<6);
	while(true){
		switch((transfer=xread_u8(&stdio_fh))>>3){
			case 0:
				if(transfer&7)proterr();
				break;
			case 1:
				if(transfer&1)xread_u64(&stdio_fh);
				if(transfer&2)xread(&stdio_fh,remote_chunk_digest,DIGEST_SALT_SIZE);
				if(transfer&4)xread(&stdio_fh,remote_chunk_digest,DIGEST_SIZE);
				break;
			case 2:
				if(transfer&7)proterr();
				return;
			default:
				proterr();
		}
	}
}

void sender_forward(){
	uint64_t len;
	uint8_t tmp;

	len=xread_u64(&sender_fh);
	if(len>ANYSINT_MAX(off_t))proterr();
	progress_launch(len);
	xwrite_u8(&receiver_fh,0);
	while(true){
		switch((tmp=xread_u8(&sender_fh))>>6){
			case 0:
				xwrite_u8(&receiver_fh,tmp);
				progress_update(false);
				break;
			case 1:
				len=xread_u64(&sender_fh);
				if(len<1||len>ANYSINT_MAX(off_t))proterr();
				xwrite_u8(&receiver_fh,1<<6);
				xwrite_u64(&receiver_fh,len);
				xcopy(&receiver_fh,&sender_fh,DIGEST_SALT_SIZE+DIGEST_SIZE);
				xcopy(&receiver_fh,&sender_fh,len);
				progress_update(true);
				break;
			case 2:
				xwrite_u8(&receiver_fh,2<<6);
				progress_stop();
				return;
			default:
				proterr();
		}
	}
}

